import { AdEntity } from '@modules/ad/core/domain/ad.entity';
import { Frequency, Role } from '@modules/ad/core/domain/ad.types';
import { CandidateEntity } from '@modules/ad/core/domain/candidate.entity';
import { DateInterval } from '../../../../domain/candidate.types';
import { Point } from '../../../types/point.type';
import { Waypoint } from '../../../types/waypoint.type';
import { Selector } from '../algorithm.abstract';
import { ScheduleItem } from '../match.query';

/**
 * This class complements the AdRepository prisma service by turning a match query object into a SQL query,
 * with the assumption that the query is passenger-oriented (i.e. it is up to the driver to go out of his way to pick up the passenger).
 * The idea is to make a rough filter of the ads in DB to limit the number of ads to be processed more precisely by the application code.
 * TODO: Converting the query object into a SQL query is a job for the repository implementation
 *       (or anything behind the repository interface),
 *       any logic related to being passenger-oriented should be in the domain layer.
 *       (though it might be difficult to describe generically the search criteria with a query object)
 */
export class PassengerOrientedSelector extends Selector {
  select = async (): Promise<CandidateEntity[]> => {
    const queryStringRoles: QueryStringRole[] = [];
    if (this.query.driver)
      queryStringRoles.push({
        query: this._createQueryString(Role.DRIVER),
        role: Role.DRIVER,
      });
    if (this.query.passenger)
      queryStringRoles.push({
        query: this._createQueryString(Role.PASSENGER),
        role: Role.PASSENGER,
      });

    return (
      await Promise.all(
        queryStringRoles.map<Promise<AdsRole>>(
          async (queryStringRole: QueryStringRole) =>
            <AdsRole>{
              ads: await this.repository.getCandidateAds(queryStringRole.query),
              role: queryStringRole.role,
            },
        ),
      )
    )
      .map((adsRole: AdsRole) =>
        adsRole.ads.map((adEntity: AdEntity) =>
          CandidateEntity.create({
            id: adEntity.id,
            role: adsRole.role == Role.DRIVER ? Role.PASSENGER : Role.DRIVER,
            frequency: adEntity.getProps().frequency,
            dateInterval: this._fixDateInterval({
              lowerDate: this._maxDateString(
                this.query.fromDate,
                adEntity.getProps().fromDate,
              ),
              higherDate: this._minDateString(
                this.query.toDate,
                adEntity.getProps().toDate,
              ),
            }),
            driverWaypoints:
              adsRole.role == Role.PASSENGER
                ? adEntity.getProps().waypoints
                : this.query.waypoints.map((waypoint: Waypoint) => ({
                    position: waypoint.position,
                    lon: waypoint.lon,
                    lat: waypoint.lat,
                  })),
            passengerWaypoints:
              adsRole.role == Role.DRIVER
                ? adEntity.getProps().waypoints
                : this.query.waypoints.map((waypoint: Waypoint) => ({
                    position: waypoint.position,
                    lon: waypoint.lon,
                    lat: waypoint.lat,
                  })),
            driverDistance:
              adsRole.role == Role.PASSENGER
                ? (adEntity.getProps().driverDistance as number)
                : (this.query.driverRoute?.distance as number),
            driverDuration:
              adsRole.role == Role.PASSENGER
                ? (adEntity.getProps().driverDuration as number)
                : (this.query.driverRoute?.duration as number),
            driverSchedule:
              adsRole.role == Role.PASSENGER
                ? adEntity.getProps().schedule
                : this.query.schedule?.map((scheduleItem: ScheduleItem) => ({
                    day: scheduleItem.day as number,
                    time: scheduleItem.time,
                    margin: scheduleItem.margin as number,
                  })),
            passengerSchedule:
              adsRole.role == Role.DRIVER
                ? adEntity.getProps().schedule
                : this.query.schedule?.map((scheduleItem: ScheduleItem) => ({
                    day: scheduleItem.day as number,
                    time: scheduleItem.time,
                    margin: scheduleItem.margin as number,
                  })),
            spacetimeDetourRatio: {
              maxDistanceDetourRatio: this.query
                .maxDetourDistanceRatio as number,
              maxDurationDetourRatio: this.query
                .maxDetourDurationRatio as number,
            },
          }),
        ),
      )
      .flat();
  };

  private _createQueryString = (role: Role): string =>
    [
      this._createSelect(role),
      this._createFrom(),
      'WHERE',
      this._createWhere(role),
    ]
      .join(' ')
      .replace(/\s+/g, ' '); // remove duplicate spaces for easy debug !

  private _createSelect = (role: Role): string =>
    [
      `SELECT \
        ad.uuid,driver,passenger,frequency,public.st_astext(ad.waypoints) as waypoints,\
        "fromDate","toDate",\
        "seatsProposed","seatsRequested",\
        strict,\
        "fwdAzimuth","backAzimuth",\
        ad."createdAt",ad."updatedAt",\
        si.uuid as "scheduleItemUuid",si.day,si.time,si.margin,si."createdAt" as "scheduleItemCreatedAt",si."updatedAt" as "scheduleItemUpdatedAt"`,
      role == Role.DRIVER ? this._selectAsDriver() : this._selectAsPassenger(),
    ].join();

  private _selectAsDriver = (): string =>
    `${this.query.driverRoute?.duration} as driverDuration,${this.query.driverRoute?.distance} as driverDistance`;

  private _selectAsPassenger = (): string =>
    `"driverDuration","driverDistance"`;

  private _createFrom = (): string =>
    'FROM ad LEFT JOIN schedule_item si ON ad.uuid = si."adUuid"';

  private _createWhere = (role: Role): string =>
    [
      this._whereRole(role),
      this._whereStrict(),
      this._whereDate(role),
      this._whereExcludedAd(),
      this._whereAzimuth(),
      this._whereProportion(role),
      this._whereRemoteness(role),
    ]
      .filter((where: string) => where != '')
      .join(' AND ');

  private _whereRole = (role: Role): string =>
    role == Role.PASSENGER ? 'driver=True' : 'passenger=True';

  private _whereStrict = (): string =>
    this.query.strict
      ? this.query.frequency == Frequency.PUNCTUAL
        ? `frequency='${Frequency.PUNCTUAL}'`
        : `frequency='${Frequency.RECURRENT}'`
      : '';

  /**
   * Generates the WHERE clause checking that the date range of the query intersects with the range of the ad.
   * Note that driver dates might not be comparable with passenger dates when the trip is by night or very long.
   * For this reason, the pickup date is adjusted with the driver duration,
   * so as to compare with the maximum / minimum driver date that could make sense for the passenger.
   * This may return more ads than necessary, but they will be filtered out in further processing.
   */
  private _whereDate = (role: Role): string => {
    const maxFromDate = this._maxFromDate(role);
    const minToDate = this._minToDate(role);
    return `("fromDate" <= ${maxFromDate} AND "toDate" >= ${minToDate})`;
  };

  private _maxFromDate = (role: Role): string => {
    if (role == Role.DRIVER) {
      //When looking for a passenger, we add the duration of the driver route to the latest toDate
      //to compute the maximum sensible passenger fromDate, in case the pickup date could be on the next day
      const querySchedule = this.query.schedule;
      // When there is no schedule (search whole day), we consider the driver accepts to depart until 23:59
      const maxScheduleTime =
        querySchedule === undefined
          ? '23:59'
          : querySchedule.reduce(
              (max, s) => (s.time > max ? s.time : max),
              '00:00',
            );
      const [h, m] = maxScheduleTime.split(':');
      const maxFromDate = new Date(this.query.toDate);
      maxFromDate.setHours(parseInt(h));
      maxFromDate.setMinutes(parseInt(m));
      maxFromDate.setSeconds(this.query.driverRoute!.duration);
      return `'${maxFromDate.getUTCFullYear()}-${maxFromDate.getUTCMonth() + 1}-${maxFromDate.getUTCDate()}'`;
    } else {
      return `'${this.query.toDate}'`;
    }
  };

  private _minToDate = (role: Role): string => {
    if (role == Role.PASSENGER) {
      // When looking for a driver, we look for a toDate that is one day before the fromDate of the query
      // so that the driver will be able to pick up the passenger even during a long trip that starts the day before
      const oneDayBeforeFromDate = new Date(this.query.fromDate);
      oneDayBeforeFromDate.setDate(oneDayBeforeFromDate.getDate() - 1);
      return `'${oneDayBeforeFromDate.getUTCFullYear()}-${oneDayBeforeFromDate.getUTCMonth() + 1}-${oneDayBeforeFromDate.getUTCDate()}'`;
    } else {
      return `'${this.query.fromDate}'`;
    }
  };

  private _whereExcludedAd = (): string =>
    this.query.excludedAdId ? `ad.uuid <> '${this.query.excludedAdId}'` : '';

  private _whereAzimuth = (): string => {
    if (!this.query.useAzimuth) return '';
    const { minAzimuth, maxAzimuth } = this._azimuthRange(
      this.query.backAzimuth as number,
      this.query.azimuthMargin as number,
    );
    if (minAzimuth <= maxAzimuth)
      return `("fwdAzimuth" <= ${minAzimuth} OR "fwdAzimuth" >= ${maxAzimuth})`;
    return `("fwdAzimuth" <= ${minAzimuth} AND "fwdAzimuth" >= ${maxAzimuth})`;
  };

  private _whereProportion = (role: Role): string => {
    if (!this.query.useProportion) return '';
    switch (role) {
      case Role.PASSENGER:
        return `(${this.query.passengerRoute?.distance}>(${this.query.proportion}*"driverDistance"))`;
      case Role.DRIVER:
        return `("passengerDistance">(${this.query.proportion}*${this.query.driverRoute?.distance}))`;
    }
  };

  private _whereRemoteness = (role: Role): string => {
    this.query.waypoints.sort(
      (firstWaypoint: Waypoint, secondWaypoint: Waypoint) =>
        firstWaypoint.position - secondWaypoint.position,
    );
    switch (role) {
      case Role.PASSENGER:
        return `\
        public.st_distance('POINT(${this.query.waypoints[0].lon} ${
          this.query.waypoints[0].lat
        })'::public.geography,direction)<\
        ${this.query.remoteness} AND \
        public.st_distance('POINT(${
          this.query.waypoints[this.query.waypoints.length - 1].lon
        } ${
          this.query.waypoints[this.query.waypoints.length - 1].lat
        })'::public.geography,direction)<\
        ${this.query.remoteness}`;
      case Role.DRIVER:
        const lineStringPoints: string[] = [];
        this.query.driverRoute?.points.forEach((point: Point) =>
          lineStringPoints.push(
            `public.st_makepoint(${point.lon},${point.lat})`,
          ),
        );
        const lineString = [
          'public.st_makeline( ARRAY[ ',
          lineStringPoints.join(','),
          '] )::public.geography',
        ].join('');
        return `\
        public.st_distance( public.st_startpoint(waypoints::public.geometry), ${lineString})<\
        ${this.query.remoteness} AND \
        public.st_distance( public.st_endpoint(waypoints::public.geometry), ${lineString})<\
        ${this.query.remoteness}`;
    }
  };

  private _azimuthRange = (
    azimuth: number,
    margin: number,
  ): { minAzimuth: number; maxAzimuth: number } => ({
    minAzimuth:
      azimuth - margin < 0 ? azimuth - margin + 360 : azimuth - margin,
    maxAzimuth:
      azimuth + margin > 360 ? azimuth + margin - 360 : azimuth + margin,
  });

  //TODO If the dates are always formatted with '%Y-%m-%d', no conversion to Date is needed
  private _maxDateString = (date1: string, date2: string): string =>
    new Date(date1) > new Date(date2) ? date1 : date2;

  private _minDateString = (date1: string, date2: string): string =>
    new Date(date1) < new Date(date2) ? date1 : date2;

  /**
   * When a punctual ad matches a punctual query, it may be on a different date than the query
   * (for routes by night), and the range produced by _minDateString and _maxDateString is not correct.
   * This function fixes that by inverting the dates if necessary.
   */
  private _fixDateInterval(interval: DateInterval): DateInterval {
    if (interval.lowerDate > interval.higherDate) {
      const tmp = interval.lowerDate;
      interval.lowerDate = interval.higherDate;
      interval.higherDate = tmp;
    }
    return interval;
  }
}

export type QueryStringRole = {
  query: string;
  role: Role;
};

type AdsRole = {
  ads: AdEntity[];
  role: Role;
};
