import { AdEntity } from './ad.entity';
import { Role, UserAd } from './ad.types';
import { GeorouterService } from './georouter.service';
import {
  Path,
  PathCreator,
  PathType,
  TypedRoute,
} from './path-creator.service';
import { Point } from './value-objects/point.value-object';

export class AdFactory {
  constructor(private readonly routeProvider: GeorouterService) {}
  /**
   * Create an AdEntity (a "matcher ad", that is: the data needed to match an ad with a match query)
   * from a "user ad" (the data provided by the user).
   */
  public async create(ad: UserAd): Promise<AdEntity> {
    const roles: Role[] = [];
    if (ad.driver) roles.push(Role.DRIVER);
    if (ad.passenger) roles.push(Role.PASSENGER);

    const pathCreator = new PathCreator(
      roles,
      ad.waypoints.map((wp) => new Point({ lon: wp.lon, lat: wp.lat })),
    );

    let typedRoutes: TypedRoute[];
    try {
      typedRoutes = await Promise.all(
        pathCreator.getBasePaths().map(async (path: Path) => ({
          type: path.type,
          route: await this.routeProvider.getRoute({
            waypoints: path.waypoints,
          }),
        })),
      );
    } catch (e: any) {
      throw new Error('Unable to find a route for given waypoints');
    }

    let driverDistance: number | undefined;
    let driverDuration: number | undefined;
    let passengerDistance: number | undefined;
    let passengerDuration: number | undefined;
    let points: Point[];
    let fwdAzimuth: number;
    let backAzimuth: number;
    try {
      typedRoutes.forEach((typedRoute: TypedRoute) => {
        if ([PathType.DRIVER, PathType.GENERIC].includes(typedRoute.type)) {
          driverDistance = typedRoute.route.distance;
          driverDuration = typedRoute.route.duration;
          points = typedRoute.route.points.map((point) => new Point(point));
          fwdAzimuth = typedRoute.route.fwdAzimuth;
          backAzimuth = typedRoute.route.backAzimuth;
        }

        if ([PathType.PASSENGER, PathType.GENERIC].includes(typedRoute.type)) {
          passengerDistance = typedRoute.route.distance;
          passengerDuration = typedRoute.route.duration;
          if (!points) {
            points = typedRoute.route.points.map((point) => new Point(point));
          }
          if (!fwdAzimuth) fwdAzimuth = typedRoute.route.fwdAzimuth;
          if (!backAzimuth) backAzimuth = typedRoute.route.backAzimuth;
        }
      });
    } catch (error: any) {
      throw new Error('Invalid route');
    }

    return AdEntity.create({
      id: ad.id,
      driver: ad.driver,
      passenger: ad.passenger,
      frequency: ad.frequency,
      fromDate: ad.fromDate,
      toDate: ad.toDate,
      schedule: ad.schedule,
      seatsProposed: ad.seatsProposed,
      seatsRequested: ad.seatsRequested,
      strict: ad.strict,
      waypoints: ad.waypoints,
      points: points!,
      driverDistance,
      driverDuration,
      passengerDistance,
      passengerDuration,
      fwdAzimuth: fwdAzimuth!,
      backAzimuth: backAzimuth!,
    });
  }
}
