import { AggregateID, ConflictException } from '@mobicoop/ddd-library';
import {
  AD_MESSAGE_PUBLISHER,
  AD_REPOSITORY,
  AD_ROUTE_PROVIDER,
} from '@modules/ad/ad.di-tokens';
import { CreateAdCommand } from '@modules/ad/core/application/commands/create-ad/create-ad.command';
import { CreateAdService } from '@modules/ad/core/application/commands/create-ad/create-ad.service';
import { AdEntity } from '@modules/ad/core/domain/ad.entity';
import { AdAlreadyExistsException } from '@modules/ad/core/domain/ad.errors';
import { CreateAdProps, Frequency } from '@modules/ad/core/domain/ad.types';
import { GeorouterService } from '@modules/ad/core/domain/georouter.service';
import { PointProps } from '@modules/ad/core/domain/value-objects/point.value-object';
import { Test, TestingModule } from '@nestjs/testing';

const originWaypoint: PointProps = {
  lat: 48.689445,
  lon: 6.17651,
};
const destinationWaypoint: PointProps = {
  lat: 48.8566,
  lon: 2.3522,
};
const createAdProps: CreateAdProps = {
  id: '4eb6a6af-ecfd-41c3-9118-473a507014d4',
  fromDate: '2023-12-21',
  toDate: '2023-12-21',
  schedule: [
    {
      day: 4,
      time: '08:15',
      margin: 900,
    },
  ],
  driver: true,
  passenger: true,
  seatsProposed: 3,
  seatsRequested: 1,
  strict: false,
  frequency: Frequency.PUNCTUAL,
  waypoints: [originWaypoint, destinationWaypoint],
  driverDistance: 23000,
  driverDuration: 900,
  passengerDistance: 23000,
  passengerDuration: 900,
  fwdAzimuth: 283,
  backAzimuth: 93,
  points: [],
};

const mockAdRepository = {
  insertExtra: jest
    .fn()
    .mockImplementationOnce(() => ({}))
    .mockImplementationOnce(() => ({}))
    .mockImplementationOnce(() => {
      throw new Error();
    })
    .mockImplementationOnce(() => {
      throw new ConflictException('already exists');
    }),
};

const mockRouteProvider: GeorouterService = {
  getRoute: jest
    .fn()
    .mockImplementationOnce(() => {
      throw new Error();
    })
    .mockImplementationOnce(() => ({
      distance: 350101,
      duration: 14422,
      fwdAzimuth: 273,
      backAzimuth: 93,
      distanceAzimuth: 336544,
      points: undefined,
    }))
    .mockImplementation(() => ({
      distance: 350101,
      duration: 14422,
      fwdAzimuth: 273,
      backAzimuth: 93,
      distanceAzimuth: 336544,
      points: [
        {
          lon: 6.1765102,
          lat: 48.689445,
        },
        {
          lon: 4.984578,
          lat: 48.725687,
        },
        {
          lon: 2.3522,
          lat: 48.8566,
        },
      ],
    })),
};

const mockMessagePublisher = {
  publish: jest.fn().mockImplementation(),
};

describe('create-ad.service', () => {
  let createAdService: CreateAdService;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: AD_REPOSITORY,
          useValue: mockAdRepository,
        },
        {
          provide: AD_ROUTE_PROVIDER,
          useValue: mockRouteProvider,
        },
        {
          provide: AD_MESSAGE_PUBLISHER,
          useValue: mockMessagePublisher,
        },
        CreateAdService,
      ],
    }).compile();

    createAdService = module.get<CreateAdService>(CreateAdService);
  });

  it('should be defined', () => {
    expect(createAdService).toBeDefined();
  });

  describe('execution', () => {
    const createAdCommand = new CreateAdCommand(createAdProps);
    it('should throw an error if route cant be computed', async () => {
      await expect(
        createAdService.execute(createAdCommand),
      ).rejects.toBeInstanceOf(Error);
    });
    it('should throw an error if route is corrupted', async () => {
      await expect(
        createAdService.execute(createAdCommand),
      ).rejects.toBeInstanceOf(Error);
    });
    it('should create a new ad as driver and passenger', async () => {
      AdEntity.create = jest.fn().mockReturnValue({
        id: '047a6ecf-23d4-4d3e-877c-3225d560a8da',
      });
      const result: AggregateID =
        await createAdService.execute(createAdCommand);
      expect(result).toBe('047a6ecf-23d4-4d3e-877c-3225d560a8da');
    });
    it('should create a new ad as passenger', async () => {
      AdEntity.create = jest.fn().mockReturnValue({
        id: '047a6ecf-23d4-4d3e-877c-3225d560a8da',
      });
      const result: AggregateID = await createAdService.execute({
        ...createAdCommand,
        driver: false,
      });
      expect(result).toBe('047a6ecf-23d4-4d3e-877c-3225d560a8da');
    });
    it('should throw an error if something bad happens', async () => {
      await expect(
        createAdService.execute(createAdCommand),
      ).rejects.toBeInstanceOf(Error);
    });
    it('should throw an exception if Ad already exists', async () => {
      await expect(
        createAdService.execute(createAdCommand),
      ).rejects.toBeInstanceOf(AdAlreadyExistsException);
    });
  });
});
